package geniusweb.javavenv;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.List;
import java.util.logging.Level;

import tudelft.utilities.listener.DefaultListenable;
import tudelft.utilities.logging.Reporter;

/**
 * A pipe to a python program. The python program is only created after run is
 * called, so that you can first add your listener
 */
public class Pipe extends DefaultListenable<String> {
	protected static final int MAX_BLOCK_SIZE = 20000000;
	private final List<String> allargs = new LinkedList<>();
	private final int deadlinems;

	// set when ween counter an error
	private transient Exception exception = null;
	private transient PythonError pythonerror = null;
	// set when process is started
	private transient OutputStream out = null;
	Process p = null;
	private Reporter logger;

	/**
	 * Call python with given arguments. call {@link #run()} to run.
	 * 
	 * @param pythonexe  a working python3 exe, typically
	 *                   "/usr/local/bin/python3" or something in a venv on temp
	 *                   disk.
	 * @param args       the other args for the call. The args must make up a
	 *                   command that puts the python program into a
	 *                   stdin-stdout-stderr pipe mode so that we can create a
	 *                   pipe.
	 * @param deadlinems the deadline (ms) after which the connection shoubd
	 *                   have been created. FIXME IGNORED CURRENTLY.
	 * @param logger     the {@link Reporter} for the associated party
	 */
	public Pipe(String pythonexe, List<String> args, int deadlinems,
			Reporter logger) {
		if (pythonexe == null || args == null || logger == null)
			throw new IllegalArgumentException("args must be not null");
		allargs.add(pythonexe);
		allargs.addAll(args);
		this.deadlinems = deadlinems;
		this.logger = logger;
	}

	/**
	 * Send data down the pipe
	 * 
	 * @param data the data to send
	 * @throws IOException if the pipe broke
	 */
	public void send(String data) throws IOException {
		// System.out.println("Sending " + data);
		byte[] lenbuf = new byte[4];
		ByteBuffer bb = ByteBuffer.wrap(lenbuf);
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.putInt(data.length());
//		System.out.println("lenbuf=" + lenbuf[0] + "," + lenbuf[1] + ","
//				+ lenbuf[2] + "," + lenbuf[3]);

		out.write(lenbuf);
		out.write(data.getBytes("UTF8"));
		out.flush();
	}

	/**
	 * start the process and connect the pipe. Returns immediately after opening
	 * the channels to the party.
	 * 
	 * @throws IOException if we can not start up the process
	 */
	public void run() throws IOException {
		ProcessBuilder processBuilder = new ProcessBuilder(allargs);
		// System.out.println("all args " + allargs);
		p = processBuilder.start();
		out = p.getOutputStream();

		Runnable stdinreader = new Runnable() {
			@Override
			public void run() {

				String totalreply = "";
				byte[] cbuf = new byte[500];
				InputStream is = p.getInputStream();
				try {
					while (true) {
						int n = is.read(cbuf, 0, 4);
						if (n != 4)
							break;
						ByteBuffer bb = ByteBuffer.wrap(cbuf);
						bb.order(ByteOrder.LITTLE_ENDIAN);
						int remaining = bb.getInt();
						if (remaining > MAX_BLOCK_SIZE)
							throw new IOException("Stdin data exceeds "
									+ MAX_BLOCK_SIZE
									+ " bytes. Assuming corrupted input stream");
						// System.out.println("Received header for " +
						// remaining);
						if (remaining == 0)
							break;
						while (remaining > 0) {
							n = is.read(cbuf);
							if (n == -1)
								break;
							totalreply += new String(cbuf, 0, n, "UTF8");
							remaining -= n;
						}
						notifyListeners(totalreply);
						totalreply = "";
					}
				} catch (IOException e) {
					e.printStackTrace();
					if (exception == null)
						exception = e;
				}
				logger.log(Level.INFO, "Input pipe from party closed.");
				// Final null to trigger tell we ended
				notifyListeners(null);
			}
		};

		/**
		 * stderr is not so strictly controlled by us. Python may print at any
		 * time to this. We assume that the party will exit shortly after the
		 * exception and that the pipe will break then. We just accumulate and
		 * store the error
		 */
		Runnable errorlogger = new Runnable() {
			@Override
			public void run() {
				byte[] cbuf = new byte[500];
				InputStream is = p.getErrorStream();
				String totalerror = "";
				try {
					while (true) {
						int n = is.read(cbuf);
						if (n == -1)
							break;
						String str = new String(cbuf, 0, n, "UTF8");
						logger.log(Level.WARNING, str);
						totalerror += str;
					}
					if (!totalerror.isEmpty())
						pythonerror = new PythonError(totalerror);
				} catch (IOException e) {
					if (exception == null)
						exception = e;
				}
			}

		};

		new Thread(stdinreader).start();
		new Thread(errorlogger).start();
	}

	/**
	 * 
	 * @return an error, either picked up from stderr (if the party prints some
	 *         error), or if that is null, from our own exception handling.
	 *         Returns null if both are not set.
	 */
	public Exception getError() {
		if (pythonerror != null)
			return pythonerror;
		return exception;
	}

	/**
	 * Terminate the pipes. Kill the party.
	 */
	public void terminate() {
		if (p != null) {
			p.destroyForcibly();
			// that should also terminate the sockets.
			p = null;
		}
	}
}
